/*
 * ListEdit.h - ListEdit header file
 */

#ifndef	_ListEdit_

# define	_ListEdit_


# ifndef	New
# define	New(x)		(x **) NewHandle ((Size) sizeof (x))
# endif


/*
 * A LineList object consists of the following:
 *
 * o	owning port
 * o	scroll bar
 * o	rectangle indicating the text display area
 * o	maximum number of lines
 * o	number of lines visible in text display area
 * o	index of top visible line
 * o	height of each line in the display
 * o	current number of lines
 * o	currently selected line (noLine if none)
 * o	activation state
 * o	number of fields in each line
 * o	array indicating horizontal offset of each field
 * o	array of handles to information for each line.
 *
 * Lines are lists of fields.
 * A field consists of a string and a handle to the next field.
 * A LineHandle is equivalent to a FieldHandle because a line is
 * simply a list of fields.
 */

# define	noLine		(-1)	/* "current line" value of no line selected */


typedef struct Field
{
	Str255			fStr;
	struct Field	**fNext;
} Field, **FieldHandle, **LineHandle;


typedef LineHandle	LineArray[1];
typedef LineArray	**LAHandle;


typedef struct LineList
{
	GrafPtr			port;		/* port                          */
	ControlHandle	scroll;		/* scroll bar                    */
	Rect			textRect;	/* text display rectangle        */
	short			maxLines;	/* maximum number of lines       */
	short			visLines;	/* number of visible lines       */
	short			topVisLine;	/* index of top visible line     */
	short			lineHeight;	/* height of each line           */
	short			nLines;		/* number of lines               */
	short			curLine;	/* currently selected line       */
	Boolean			hilite;		/* whether to hilite             */
	short			nFields;	/* number of fields in each line */
	short			*offset;	/* offsets of each field         */
	LAHandle		lines;		/* lines in list                 */
} LineList, *LListPtr;


# define	NextField(hField)	((**hField).fNext)
# define	ListLine(list, line)	((**list->lines)[line])

void InitList (LListPtr theList, short maxLines);
LineHandle NewLine (short nFields);
void DisposeLine (LListPtr theList, short n);
void SetFieldStr (FieldHandle field, StringPtr str);
short FindLine (LListPtr theList, Point pt);
void ScrollToLine (LListPtr theList, short n);
void ListScrollRel (LListPtr theList, short n);
void ListSetScroll (LListPtr theList);
void ListScrollMax (LListPtr theList);
void SelectLine (LListPtr theList, short n);
Boolean ListTestScroll (LListPtr theList, Point pt);
Boolean ListTestText (LListPtr theList, Point pt);
Boolean InsertLine (LListPtr theList, LineHandle theLine, short n);
void DeleteLine (LListPtr theList, short n);
void PasteField (LListPtr theList, short lineNo, short fieldNo, StringPtr str);
void PasteLine (LListPtr theList, LineHandle theLine, short n);
void ListActivate (LListPtr theList, Boolean hilite);
void ResetList (LListPtr theList);
void GetLineRect (Rect *r, LListPtr theList, short lineNo);
void GetFieldRect (Rect *r, LListPtr theList, short lineNo, short fieldNo);
void ScrollListLines (LListPtr theList, short from, short to);
Boolean LineVisible (LListPtr theList, short lineNo);
void InvertLine (LListPtr theList, short lineNo);
void DrawField (LListPtr theList, short lineNo, short fieldNo);
void DrawLine (LListPtr theList, short lineNo);
void DrawListLines (LListPtr theList, short from, short to);
void DrawListText (LListPtr theList);
void DrawListFrame (LListPtr theList);
void DrawList (LListPtr theList);

# endif /* _ListEdit_ */
